;;;; delimited-control.scm
;;;; Kon Lovett, Apr 6 '06

;; prompt, prompt0, reset, reset0 are synonyms


;FIXME single-threaded

(module delimited-control

(;export
  (prompt $prompt) (prompt0 $prompt)
  (reset $prompt) (reset0 $prompt)
  (control $control) (control-abort $control)
  (control0 $control)
  (shift $control)
  (shift0 $control))

(import scheme
  (chicken base)
  (chicken type)
  (only miscmacros let/cc))

(: $prompt (procedure -> . *))
(: $control (boolean boolean procedure -> . *))

;;;

;; prompt-control

;

(define-syntax prompt
	(syntax-rules ()
		((_ EXPR) ($prompt (lambda () EXPR))) ) )

(define-syntax prompt0
	(syntax-rules ()
		((_ EXPR) ($prompt (lambda () EXPR))) ) )

; +F-

(define-syntax control
	(syntax-rules ()
		((_ PP EXPR) ($control #f #t (lambda (PP) EXPR))) ) )

(define-syntax control-abort
	(syntax-rules ()
		((_ V) ($control #f #t (lambda (_) V))) ) )

; -F-

(define-syntax control0
	(syntax-rules ()
		((_ PP EXPR) ($control #f #f (lambda (PP) EXPR))) ) )

;; reset-shift

;

(define-syntax reset
	(syntax-rules ()
		((_ EXPR) ($prompt (lambda () EXPR))) ) )

(define-syntax reset0
	(syntax-rules ()
		((_ EXPR) ($prompt (lambda () EXPR))) ) )

; +F+

(define-syntax shift
	(syntax-rules ()
		((_ PP EXPR) ($control #t #t (lambda (PP) EXPR))) ) )

; -F+

(define-syntax shift0
	(syntax-rules ()
		((_ PP EXPR) ($control #t #f (lambda (PP) EXPR))) ) )

;;
;; Local impl of continuation binding
;;

#;
(define-syntax let/ccp
  (syntax-rules ()
    ((_ K BODY ...) (continuation-capture (lambda (K) BODY ...)) ) ) )

;; This is one single global mutable cell, holding the stack of holes
;;
;; A hole has a continuation and a mark. The mark, if #t, says if the
;; hole is a delimiting hole. A non-delimiting hole is just like the
;; return from a regular function.

(define *holes* (make-parameter '()))

(define (hole-push! hole) (*holes* (cons hole (*holes*))))

(define (hole-list-push! holes-org)
  (when (pair? holes-org)
    (let loop ((holes holes-org))
      (let ((next-hole (cdr holes)))
        (if (not (null? next-hole)) (loop next-hole)
            (begin
              (set-cdr! holes (*holes*))
              (*holes* holes-org) ) ) ) ) ) )

(define (hole-pop!*)
	(let ((holes (*holes*)))
		(*holes* (cdr holes))
		(car holes) ) )

(define (hole-pop!)
  (if (not (null? (*holes*))) (hole-pop!*)
      (error 'control "missing toplevel delimited control operator") ) )

;;

(define (cell-new v mark) (cons v mark))
(define (cell-ref c) (car c))
(define (cell-marked? c) (cdr c))

;;

(define (unwind-till-marked! keep-delimited?)
  (let loop ((hole (hole-pop!)) (holes-prefix '()))
    (if (not (cell-marked? hole))
        ;then mark (?)
        (loop (hole-pop!) (cons hole holes-prefix))
        ;else marked, it's prompt's hole
        (begin
          ;put it back
          (hole-push! (if keep-delimited? hole
                          ;make the hole non-delimiting
                          (cell-new (cell-ref hole) #f)))
          holes-prefix ) ) ) )

;; Essentially this is the ``return from the function''

(: *abort-top! (* -> . *))

#;
(define (*abort-top! v)
  ((cell-ref (hole-pop!*)) v) )

(define-syntax *abort-top!
	(syntax-rules ()
		((*abort-top! EXPR)
      (call-with-values
        (lambda () EXPR)
        (lambda vals
          (apply (cell-ref (hole-pop!*)) vals))) ) ) )

(define ($prompt thunk)
	(let/cc outer-k
		(hole-push! (cell-new outer-k #t)) ; it's prompt's hole
		(*abort-top! (thunk)) ) )

(define ($control shift? keep-delimited? func)
  (let/cc k-control
		(let* ((holes-prefix (unwind-till-marked! keep-delimited?))
					 (invoke-subcont (lambda vals
                             (let/cc k-return
                               (hole-push! (cell-new k-return shift?))
                               (hole-list-push! holes-prefix)
                               (apply k-control vals)))) )
			(*abort-top! (func invoke-subcont)) ) ) )

) ;module delimited-control
